/* COMP_SM.v
*******************************************************************************
 Actel Corportation
 Fusion's Voltage Comparator Application Note's Design Example

 File Name:    TOP_CM.vhd
 Date:         December 2006
 Language:     Verilog

*******************************************************************************
 IMPORTANT-READ THESE TERMS CAREFULLY BEFORE UTILIZING THE INFORMATION
 CONTAINED IN THIS APPLICATION NOTE. 
 NO SUPPORT/NO WARRANTY. THE DESIGN FILES AND DOCUMENTATION ARE PROVIDED
 WITHOUT SUPPORT OR WARRANTY OF ANY KIND. ACTEL IS NOT OBLIGATED TO 
 PROVIDE UPDATES, BUG FIXES, OR TECHNICAL SUPPORT AND DISCLAIMS ALL 
 WARRANTIES INCLUDING, WITHOUT LIMITATION, THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE, AND WARRANTIES OF NON-INFRINGEMENT OF THE
 RIGHTS OF THIRD PARTIES (INCLUDING, WITHOUT LIMITATION, RIGHTS UNDER PATENT,
 COPYRIGHT, TRADE SECRET, OR OTHER INTELLECTUAL PROPERTY RIGHTS).  RECIPIENT
 ACCEPTS THE APPLICATION NOTE DESIGN FILES AND DOCUMENTATION IN "AS-IS"
 CONDITION.
*******************************************************************************
***************************************************************************************************
 Main Characteristics and Requirements:
  The state machine supports 6 comparator measurement channels (using 12 analog
  voltage inputs). If more channels are desired, the state machine I/Os 
  should be modified accordingly. The measurement resolution is set to 12-bit as default but
  it can be changed using `WIDTH generic.
  Each comparator channel is comprised of two voltage inputs (P and N ports). The
  state machine does not restrict the sampling sequence of AB but requires that the 'N' voltage
  of a channel be measured immediately after 'P' voltage of that channel. If it is not, the measurement
  for this channel will be invalidated.
***************************************************************************************************
 Functional Description:
  The state machine starts off in 'idle' state. In this state all the measurement results are
  reset to zero. The latest_update output is set to "111". The state machine keeps looking at 
  the channels measured by AB (monitoring ASSC_DONE and CHNUMBER). Once the AB meausures the P voltage 
  of any comparator channels, the state machine moves to channel_xP state. At this state the 
  measurement results are recorded and state machine moves to 'wait_next_measurement' state.
  The state machine expects the next measurement to be taken from the N port of the same channel. If the
  next measurement is not from the expected channel, it will go to idle state, invalidating the
  previous P measurement and also clearing all the output results to signal an error in sampling sequence. 
  If the correct N voltage is measured after P, the state machine moves to "channel_xN" state in which
  the results are recorded. This result along with the corresponding P result are entered into the subtractor 
  and the output of the subtractor is sent to the corresponding Diff channel output of the state machine.
  Then the state machine will move to "wait_next_channel" state and wait for next Dcomparator measurement.
****************************************************************************************************/

`define WIDTH 12

module COMP_SM (SYSCLK, RESETn, DATAVALID, INIT_DONE, ADC_CHNUMBER, ADC_RESULT, 
                Diff_0, Diff_1, Diff_2, Diff_3, Diff_4, Diff_5, Latest_Update,
                New_Measurement);
//defparam Subtract.WIDTH_SUB = 12;
//parameter WIDTH = 12;

//defining states of the state machine
 parameter idle=3'b000, channel_xP=3'b001, channel_xN=3'b010, wait_next_result=3'b011, 
           wait_next_channel=3'b100;

 parameter AV0 = 5'b00001; 
 parameter AC0 = 5'b00010; 
 parameter AV1 = 5'b00100; 
 parameter AC1 = 5'b00101; 
 parameter AV2 = 5'b00111; 
 parameter AC2 = 5'b01000; 
 parameter AV3 = 5'b01010; 
 parameter AC3 = 5'b01011; 
 parameter AV4 = 5'b01101; 
 parameter AC4 = 5'b01110; 
 parameter AV5 = 5'b10000; 
 parameter AC5 = 5'b10001; 
 parameter AV6 = 5'b10011; 
 parameter AC6 = 5'b10100; 
 parameter AV7 = 5'b10110; 
 parameter AC7 = 5'b10111; 
 parameter AV8 = 5'b11001; 
 parameter AC8 = 5'b11010; 
 parameter AV9 = 5'b11100; 
 parameter AC9 = 5'b11101; 

 parameter Channel0 = 3'b000;
 parameter Channel1 = 3'b001;
 parameter Channel2 = 3'b010;
 parameter Channel3 = 3'b011;
 parameter Channel4 = 3'b100;
 parameter Channel5 = 3'b101;
 parameter no_output = 3'b111; 

input SYSCLK, RESETn; // system clock and active-low reset
input DATAVALID; //driven by ASSC_DONE output of Analog Block
input INIT_DONE; // driven by INIT_DONE output of of Analog Block
input [4:0] ADC_CHNUMBER; //driven by ADC_CHNUMBER of Analog Block
input [`WIDTH-1:0] ADC_RESULT; //driven by ADC_RESULT of Analog Block
output [`WIDTH-1:0] Diff_0, Diff_1, Diff_2, Diff_3, Diff_4, Diff_5; //6 Diff measurement channels outputs
output [2:0] Latest_Update; //indicates the Diff measurement channel measured last
output New_Measurement; // active high for one clock period. Indicates outputing a new Diff measurement

reg [`WIDTH-1:0] Diff_0, Diff_1, Diff_2, Diff_3, Diff_4, Diff_5;
reg [2:0] Latest_Update;
reg New_Measurement;

 //Other Signal Definitions
 /********************************************************************/
 // Pair of Registers holding the measured P and N voltages of a channel
 reg [`WIDTH-1:0] R1, R2;
 // output of the subtractor (R1-R2)
wire [`WIDTH-1:0] Result;
 // Used to create New_Measurement output Flag
 reg completed, completed_sync;
 // used as enables for R1 and R2
 reg P, N;  
 // Register holding previously measured analog voltage channel
 reg [4:0] prev_channel;
 // Flag indicating if corresponding N voltage is measured immediately after P voltage
 reg valid;
 //statemachine registers
 reg [2:0] pres_state, next_state;
 /********************************************************************/

// assign next state value to presenet state register
always @ (posedge SYSCLK or negedge RESETn or negedge INIT_DONE)
 begin
 if (RESETn == 1'b0 || INIT_DONE == 1'b0)
    pres_state <= idle;
 else 
    pres_state <= next_state;
 end

// Body of the state machine defining state transitions 
always @ (pres_state or DATAVALID or ADC_CHNUMBER or valid)
begin: fsm
case (pres_state)
 idle:   case (DATAVALID)
           1'b1:  case (ADC_CHNUMBER)
                    AV0: next_state = channel_xP;
                    AV1: next_state = channel_xP;
                    AV2: next_state = channel_xP;
                    AV3: next_state = channel_xP;
                    AV4: next_state = channel_xP;
                    AV5: next_state = channel_xP;
                    default: next_state = idle;
                  endcase    
           default: next_state = idle; 
         endcase
channel_xP: next_state = wait_next_result;
wait_next_result: case (DATAVALID)
                    1'b1:  case (valid)
                             1'b1:  next_state = channel_xN;
                             default: next_state = idle;
                             endcase  
                    default: next_state = wait_next_result;
                  endcase
channel_xN: next_state = wait_next_channel;
wait_next_channel:  case (DATAVALID) 
                      1'b1:  case (ADC_CHNUMBER)
                                AV0: next_state = channel_xP;
                                AV1: next_state = channel_xP;
                                AV2: next_state = channel_xP;
                                AV3: next_state = channel_xP;
                                AV4: next_state = channel_xP;
                                AV5: next_state = channel_xP;
                                default: next_state = wait_next_channel;
                             endcase
                      default: next_state = wait_next_channel;
                    endcase
default: next_state = idle;
endcase
end //fsm
 
// output of the state machine is P and N enables to R1 and R2 to record voltage measurements.
// State machine also outputs a 'completed' signal used for creation of 'New_Measurement' flag.
always @ (pres_state) begin
 case (pres_state)
   channel_xP: begin
               P = 1'b1;
               completed = 1'b0; 
	           N = 1'b0;
			   end
   channel_xN: begin
               P = 1'b0;
               completed = 1'b1; 
               N = 1'b1;
			   end
   default:   begin
               P =  1'b0;
               N =  1'b0;
               completed = 1'b0; 
			   end
 endcase
end 

/// Assgining output of the subtractor (result) to corresponding Diff channel
always @ (posedge SYSCLK or negedge RESETn) begin 
  if (RESETn == 1'b0) 
   begin
    Diff_0 <= 0;
    Diff_1 <= 0;
    Diff_2 <= 0;
    Diff_3 <= 0;
    Diff_4 <= 0;
    Diff_5 <= 0;
	Latest_Update <= no_output;
  end
  else 
  begin
      if (completed_sync == 1'b1) 
        case (ADC_CHNUMBER) 
           AC0: begin
              Diff_0 <= Result;
              Latest_Update <= Channel0;
			   end
           AC1: begin
              Diff_1 <= Result;
              Latest_Update <= Channel1;
			   end
           AC2: begin
              Diff_2 <= Result;
              Latest_Update <= Channel2;
			   end
           AC3: begin
              Diff_3 <= Result;
              Latest_Update <= Channel3;
			   end
           AC4: begin
              Diff_4 <= Result;
              Latest_Update <= Channel4;
			    end
           AC5: begin
              Diff_5 <= Result;
              Latest_Update <= Channel5;
			   end
           default: begin
		            Diff_0 <= Diff_0;
                    Diff_1 <= Diff_0;
                    Diff_2 <= Diff_0;
                    Diff_3 <= Diff_0;
                    Diff_4 <= Diff_0;
                    Diff_5 <= Diff_0;
                    Latest_Update <= Latest_Update;
					end
         endcase
 end
 end 
// asynchrnous subtractor
Subtract #(`WIDTH) Subtract_inst(.DataA(R1), .DataB(R2), .Sum(Result)) ;
//Subtract Subtract_inst(.DataA(R1), .DataB(R2), .Sum(Result)) ;

// Recording P voltage measurements into R1 register
always @ (posedge SYSCLK or negedge RESETn) begin
 if (RESETn == 1'b0) 
   R1 <= 0;
 else
   begin
    if (P == 1'b1) 
    R1 <= ADC_RESULT;
   end
end

// Recording N voltage measurements into R2 register
always @ (posedge SYSCLK or negedge RESETn) begin
 if (RESETn == 1'b0) 
   R2 <= 0;
 else
  begin
   if (N == 1'b1) 
   R2 <= ADC_RESULT;
   end
end

// generating New measurement through two back to back registers
always @ (posedge SYSCLK or negedge RESETn) begin
 if (RESETn == 1'b0) 
   completed_sync <= 1'b0;
 else 
    completed_sync <= completed;
end
    
always @ (posedge SYSCLK or negedge RESETn) begin
 if (RESETn == 1'b0) 
   New_Measurement <= 1'b0;
 else
   New_Measurement <= completed_sync;
end

// Recording previously measured analog channel
always @ (posedge SYSCLK or negedge RESETn) begin
 if (RESETn == 1'b0) 
   prev_channel <= AV0;
 else
  begin
   if (DATAVALID == 1'b1)   
    prev_channel <= ADC_CHNUMBER;
  end 
end

// Determine of the correct N voltage channel is measured after P voltage measurement, validating analog block's sequence
always @ (prev_channel or ADC_CHNUMBER or DATAVALID or RESETn or INIT_DONE) begin
if (RESETn == 1'b0 || INIT_DONE == 1'b0)  
  valid <= 1'b0;
else
  begin
  if (DATAVALID == 1'b1) 
   begin 
    case (prev_channel)
     AV0: 
      case (ADC_CHNUMBER)
        AC0: valid <= 1'b1;
        default: valid <= 1'b0;
      endcase  
     AV1: 
      case (ADC_CHNUMBER)
        AC1: valid <= 1'b1;
        default: valid <= 1'b0;
      endcase  
     AV2:
      case (ADC_CHNUMBER)
        AC2: valid <= 1'b1;
        default: valid <= 1'b0;
      endcase  
     AV3:
      case (ADC_CHNUMBER)
        AC3: valid <= 1'b1;
        default: valid <= 1'b0;
      endcase  
     AV4:
      case (ADC_CHNUMBER)
        AC4: valid <= 1'b1;
        default: valid <= 1'b0;
      endcase  
     AV5:
      case (ADC_CHNUMBER)
        AC5: valid <= 1'b1;
        default: valid <= 1'b0;
      endcase  
     default: valid <= 1'b0;
    endcase
   end
 end
end
 



endmodule

 